@@ -3,16 +3,15 @@
 # License: GPLv3
 
 '''
-UXF is a plain text human readable storage format that may serve as a
-convenient alternative to csv, ini, json, sqlite, toml, xml, or yaml.
+UXF is a plain text human readable optionally typed storage format. UXF may
+serve as a convenient alternative to csv, ini, json, sqlite, toml, xml, or
+yaml.
 
-The uxf module can be used as an executable. Run
+The uxf module can be used as an executable. To see the command line help run:
 
     python3 -m uxf -h
 
-to see the command line help.
-
-uxf's public API provides five free functions and five classes.
+uxf's public API provides six free functions and six classes.
 
     def load(filename_or_filelike): -> (data, custom_header)
     def loads(uxf_text): -> (data, custom_header)
@@ -29,13 +28,17 @@ as UXF data. The data must be a single Map, dict, List, list, or Table.
 dumps() writes the same but into a string that's returned. See the function
 docs for additional options.
 
-    def naturalize(s)
+    def naturalize(s) -> object
 
 This takes a str and returns a bool or datetime.datetime or datetime.date or
 int or float if any of these can be parsed, otherwise returns the original
 string s. This is provided as a helper function (e.g., it is used by
 uxfconvert.py).
 
+    def realize(n: float) -> str
+
+Returns a UXF-compatible, (i.e., naturalize-able) str representing the float n.
+
     class Error
 
 Used to propagate errors (and warnings if warn_is_error is True).
@@ -52,9 +55,16 @@ that also has a .comment attribute.
 
     class Table
 
-Used to store UXF Tables. A Table has a list of fieldnames and a records
-list which is a lists of lists with each sublist having the same number of
-items as the number of fieldnames. It also has a .comment attribute.
+Used to store UXF Tables. A Table has a list of fields (name, optional type;
+see below) and a records list which is a list of lists of scalars with each
+sublist having the same number of items as the number of fields. It also has
+a .comment attribute.
+
+    class Field
+
+Used to store a Table's fields. The .vtype must be one of these strs:
+'bool', 'int', 'real', 'date', 'datetime', 'str', 'bytes', or None (which
+means accept any valid type).
 
     class NTuple
 
@@ -85,10 +95,17 @@ except ImportError:
 
 __all__ = ('__version__', 'VERSION', 'load', 'loads', 'dump', 'dumps',
            'naturalize', 'List', 'Map', 'Table', 'NTuple')
-__version__ = '0.9.7' # uxf module version
+__version__ = '0.10.0' # uxf module version
 VERSION = 1.0 # uxf file format version
 
 UTF8 = 'utf-8'
+_KEY_TYPES = {'int', 'date', 'datetime', 'str', 'bytes'}
+_VALUE_TYPES = _KEY_TYPES | {'bool', 'real'}
+_ANY_VALUE_TYPES = _VALUE_TYPES | {'list', 'map', 'table', 'ntuple'}
+_BOOL_FALSE = {'no', 'false'}
+_BOOL_TRUE = {'yes', 'true'}
+_CONSTANTS = {'null'} | _BOOL_FALSE | _BOOL_TRUE
+_BAREWORDS = _ANY_VALUE_TYPES | _CONSTANTS
 
 
 def load(filename_or_filelike, *, warn_is_error=False):
@@ -355,13 +372,19 @@ class _Lexer(_ErrorMixin):
 
 
     def read_const(self):
-        match = self.match_any_of('no', 'yes', 'null', 'true', 'false')
+        match = self.match_any_of(_BAREWORDS)
         if match == 'null':
             self.add_token(_Kind.NULL)
-        elif match in {'no', 'false'}:
+        elif match in _BOOL_FALSE:
             self.add_token(_Kind.BOOL, False)
-        elif match in {'yes', 'true'}:
+        elif match in _BOOL_TRUE:
             self.add_token(_Kind.BOOL, True)
+        elif match in _ANY_VALUE_TYPES:
+            self.add_token(_Kind.TYPE, match)
+        elif match in _VALUE_TYPES:
+            self.add_token(_Kind.VALUE_TYPE, match)
+        elif match in _ANY_VALUE_TYPES:
+            self.add_token(_Kind.ANY_VALUE_TYPE, match)
         else:
             i = self.text.find('\n', self.pos)
             text = self.text[self.pos - 1:i if i > -1 else self.pos + 8]
@@ -388,11 +411,12 @@ class _Lexer(_ErrorMixin):
         self.error(error_text)
 
 
-    def match_any_of(self, *targets):
+    def match_any_of(self, targets):
         if self.at_end():
             return None
         start = self.pos - 1
-        for target in targets:
+        for target in sorted(targets, key=lambda t: (len(t), t),
+                             reverse=True):
             if self.text.startswith(target, start):
                 self.pos += len(target) - 1 # skip past target
                 return target
@@ -450,6 +474,7 @@ class _Kind(enum.Enum):
     DATE_TIME = enum.auto()
     STR = enum.auto()
     BYTES = enum.auto()
+    TYPE = enum.auto()
     EOF = enum.auto()
 
 
@@ -542,7 +567,7 @@ class NTuple:
         if len(self._items) == 2:
             if name == 'real':
                 return self._items[0]
-            elif name in {'imag', 'imaginary'}:
+            if name in {'imag', 'imaginary'}:
                 return self._items[1]
         raise AttributeError(f'{self.__class__.__name__!r} object has '
                              f'no attribute {name!r}')
@@ -566,6 +591,7 @@ class List(collections.UserList):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.comment = None
+        self.vtype = None
 
 
 class Map(collections.UserDict):
@@ -573,81 +599,125 @@ class Map(collections.UserDict):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.comment = None
+        self.ktype = None
+        self.vtype = None
+
+
+class Field:
+
+    def __init__(self, name, vtype=None):
+        self._name = name
+        self._vtype = vtype
+
+
+    @property
+    def name(self):
+        return self._name
+
+
+    @property
+    def vtype(self):
+        return self._vtype
+
+
+    @vtype.setter
+    def vtype(self, vtype):
+        if vtype is None:
+            vtype = None # This means accept any valid type
+        elif vtype in _VALUE_TYPES:
+            self._vtype = vtype
+        else:
+            types = " ".join(sorted(_VALUE_TYPES)) + ' uxf'
+            raise Error(
+                f'expected field type to be one of: {types}, got {vtype}')
 
 
 class Table:
     '''Used to store a UXF table.
 
-    A Table has a list of fieldnames and a records list which is a lists of
-    lists with each sublist having the same number of items as the number of
-    fieldnames. It also has a .comment attribute.
+    A Table has a list of fields (name, optional type) and a records list
+    which is a list of lists of scalars. with each sublist having the same
+    number of items as the number of fields. It also has a .comment
+    attribute. (Note that the lists in a Table are plain lists not UXF
+    Lists.)
+
+    The only type-safe way to add values to a table is via .append() for
+    single values or += for single values or a sequence of values.
 
     When a Table is iterated each row is returned as a namedtuple.
     '''
 
-    def __init__(self, *, name=None, fieldnames=None, records=None,
+    def __init__(self, *, name=None, fields=None, records=None,
                  comment=None):
         '''
         A Table may be created empty, e.g., Table(). However, if records is
-        not None, then both name and fieldnames must be given.
+        not None, then both the name and fields must be given.
 
         records can be a flat list of values (which will be put into a list
-        of lists with each sublist being len(fieldnames) long), or a list of
-        lists in which case each list is _assumed_ to be len(fieldnames)
+        of lists with each sublist being len(fields) long), or a list of
+        lists in which case each list is _assumed_ to be len(fields)
         long.
 
         comment is an optional str.
         '''
         self.name = name
         self._Class = None
-        self.fieldnames = [] if fieldnames is None else fieldnames
+        self.fields = [] if fields is None else fields
         self.records = []
         self.comment = comment
+        self._table_index = 1
         if records:
             if not name:
                 raise Error('can\'t create an unnamed nonempty table')
-            if not fieldnames:
+            if not fields:
                 raise Error(
-                    'can\'t create a nonempty table without field names')
-            if isinstance(records, list):
+                    'can\'t create a nonempty table without fields')
+            if isinstance(records, (list, List)):
                 if self._Class is None:
-                    self._make_class()
-                self.records = records
+                    self._make_row_class()
+                self.records = list(records)
             else:
                 for value in records:
                     self.append(value)
 
 
-    def append_fieldname(self, name):
-        self.fieldnames.append(name)
+    def append_field(self, name, vtype=None):
+        self.fields.append(Field(name, vtype))
 
 
     def append(self, value):
         if self._Class is None:
-            self._make_class()
+            self._make_row_class()
         if (not self.records or
-                len(self.records[-1]) >= len(self.fieldnames)):
+                len(self.records[-1]) >= len(self.fields)):
             self.records.append([])
-        self.records[-1].append(value)
+        index = len(self.records[-1])
+        vtype = _table_type_for_name(self.fields[index].vtype)
+        if vtype is None or isinstance(value, vtype):
+            self.records[-1].append(value)
+        else:
+            raise Error('excpected value of type {vtype}, got value '
+                        '{value!r} of type {type(value)}')
 
 
-    def _make_class(self):
+    def _make_row_class(self):
         if not self.name:
             raise Error('can\'t use an unnamed table')
-        if not self.fieldnames:
-            raise Error('can\'t create a table with no field names')
+        if not self.fields:
+            raise Error('can\'t create a table with no fields')
         self._Class = collections.namedtuple(
-            _canonicalize(self.name, 'Table'),
-            [_canonicalize(name, f'Field{i}')
-             for i, name in enumerate(self.fieldnames, 1)])
+            _canonicalize(self.name, f'Table{self._table_index}'),
+            [_canonicalize(field.name, f'Field{i}')
+             for i, field in enumerate(self.fields, 1)])
+        self._table_index += 1
 
 
     def __iadd__(self, value):
         if not self.name:
             raise Error('can\'t append to an unnamed table')
-        if not self.fieldnames:
-            raise Error('can\'t append to a table with no field names')
-        if isinstance(value, (list, tuple)): # or: collections.abc.Sequence?
+        if not self.fields:
+            raise Error('can\'t append to a table with no fields')
+        if isinstance(value, (list, List, tuple)):
             for v in value:
                 self.append(v)
         else:
@@ -657,7 +727,7 @@ class Table:
 
     def __iter__(self):
         if self._Class is None:
-            self._make_class()
+            self._make_row_class()
         for record in self.records:
             yield self._Class(*record)
 
@@ -667,13 +737,13 @@ class Table:
 
 
     def __str__(self):
-        return (f'Table {self.name!r} {self.fieldnames!r} with '
+        return (f'Table {self.name!r} {self.fields!r} with '
                 f'{len(self.records)} records #{self.comment!r}')
 
 
     def __repr__(self):
         return (f'{self.__class__.__name__}(name={self.name!r}, '
-                f'fieldnames={self.fieldnames!r}, '
+                f'fields={self.fields!r}, '
                 f'records={self.records!r}, comment={self.comment!r})')
 
 
@@ -806,10 +876,17 @@ class _Parser(_ErrorMixin):
     def _handle_field_name(self, token):
         if token.kind is _Kind.TABLE_ROWS:
             self.states[-1] = _Expect.TABLE_VALUE
+        elif token.kind is _Kind.TYPE:
+            field = self.stack[-1].fields[-1]
+            if field.vtype is None:
+                field.vtype = token.value
+            else:
+                self.error('can only provide one type for each table '
+                           f' field, got a second type {token}')
+        elif token.kind is _Kind.TABLE_FIELD_NAME:
+            self.stack[-1].append_field(token.value)
         else:
-            if token.kind is not _Kind.TABLE_FIELD_NAME:
-                self.error(f'expected table field name, got {token}')
-            self.stack[-1].append_fieldname(token.value)
+            self.error(f'expected table field, got {token}')
 
 
     def _handle_table_value(self, token):
@@ -828,6 +905,18 @@ class _Parser(_ErrorMixin):
     def _handle_map_key(self, token):
         if token.kind is _Kind.MAP_END:
             self._on_collection_end(token)
+        elif token.kind is _Kind.TYPE:
+            parent = self.stack[-1]
+            if len(parent) > 0:
+                self.error('can only provide one or two types for '
+                           f'maps, before the first value, got {token}')
+            if parent.ktype is None:
+                parent.ktype = token.value
+            elif parent.vtype is None:
+                parent.vtype = token.value
+            else:
+                self.error('can only provide key and value types for '
+                           f'maps, got a third type {token}')
         elif token.kind in {
                 _Kind.INT, _Kind.DATE, _Kind.DATE_TIME,
                 _Kind.STR, _Kind.BYTES}:
@@ -840,11 +929,11 @@ class _Parser(_ErrorMixin):
 
     def _handle_map_value(self, token):
         if self._is_collection_start(token.kind):
-            # this adds a new list, map, or table to the stack
+            # this adds a new List, Map, or Table to the stack
             self._on_collection_start(token.kind)
-            # this adds a key-value item to the map that contains the above
-            # list, map, or table, the key being the key acquired earlier,
-            # the value being the new list, map, or table
+            # this adds a key-value item to the Map that contains the above
+            # List, Map, or Table, the key being the key acquired earlier,
+            # the value being the new List, Map, or Table
             self.stack[-2][self.keys[-1]] = self.stack[-1]
         elif self._is_collection_end(token.kind):
             self.states[-1] = _Expect.MAP_KEY
@@ -857,8 +946,20 @@ class _Parser(_ErrorMixin):
 
 
     def _handle_any_value(self, token):
-        if self._is_collection_start(token.kind):
-            # this adds a new list, map, or table to the stack
+        if token.kind is _Kind.TYPE:
+            parent = self.stack[-1]
+            if not isinstance(parent, List):
+                self.error(f'unexpected type, {token}')
+            if len(parent) > 0:
+                self.error('can only provide one type for '
+                           f'lists, before the first value, got {token}')
+            if parent.vtype is None:
+                parent.vtype = token.value
+            else:
+                self.error('can only provide one type for '
+                           f'lists, got a second type {token}')
+        elif self._is_collection_start(token.kind):
+            # this adds a new List, Map, or Table to the stack
             self._on_collection_start(token.kind)
         elif self._is_collection_end(token.kind):
             self.states.pop()
@@ -988,18 +1089,27 @@ class _Writer:
     def write_list(self, item, indent=0, *, pad, map_value=False):
         tab = '' if map_value else pad * indent
         comment = getattr(item, 'comment', None)
+        vtype = getattr(item, 'vtype', None)
         if len(item) == 0:
-            self.file.write(f'{tab}[]\n' if comment is None else
-                            f'{tab}[ #<{escape(comment)}>]\n')
+            self.file.write(f'{tab}[')
+            if comment is not None:
+                self.file.write(f' #<{escape(comment)}>')
+            if vtype is not None:
+                self.file.write(f' {vtype}')
+            if comment is not None or vtype is not None:
+                self.file.write(' ')
+            self.file.write(']\n')
         else:
             self.file.write(f'{tab}[')
             if comment is not None:
                 self.file.write(f' #<{escape(comment)}>')
+            if vtype is not None:
+                self.file.write(f' {vtype}')
             indent += 1
             is_scalar = _is_scalar(item[0])
             if is_scalar:
                 kwargs = dict(indent=0, pad=' ', map_value=False)
-                if comment is not None:
+                if comment is not None or vtype is not None:
                     self.file.write(' ')
             else:
                 self.file.write('\n')
@@ -1016,21 +1126,42 @@ class _Writer:
     def write_map(self, item, indent=0, *, pad, map_value=False):
         tab = '' if map_value else pad * indent
         comment = getattr(item, 'comment', None)
+        ktype = getattr(item, 'ktype', None)
+        vtype = getattr(item, 'vtype', None)
         if len(item) == 0:
-            self.file.write(f'{tab}{{}}\n' if comment is None else
-                            f'{tab}{{ #<{escape(comment)}> }}\n')
+            self.file.write(f'{tab}{{')
+            if comment is not None:
+                self.file.write(f' #<{escape(comment)}>')
+            if ktype is not None:
+                self.file.write(f' {ktype}')
+            if vtype is not None:
+                self.file.write(f' {vtype}')
+            if (comment is not None or ktype is not None or
+                    vtype is not None):
+                self.file.write(' ')
+            self.file.write('}\n')
         elif len(item) == 1:
             self.file.write(f'{tab}{{')
             if comment is not None:
                 self.file.write(f' #<{escape(comment)}>')
+            if ktype is not None:
+                self.file.write(f' {ktype}')
+            if vtype is not None:
+                self.file.write(f' {vtype}')
             key, value = list(item.items())[0]
             self.write_scalar(key, 1, pad=' ')
             self.file.write(' ')
             self.write_value(value, 1, pad=' ', map_value=True)
             self.file.write('}\n')
         else:
-            self.file.write(f'{tab}{{\n' if comment is None else
-                            f'{tab}{{ #<{escape(comment)}>\n')
+            self.file.write(f'{tab}{{')
+            if comment is not None:
+                self.file.write(f' #<{escape(comment)}>')
+            if ktype is not None:
+                self.file.write(f' {ktype}')
+            if vtype is not None:
+                self.file.write(f' {vtype}')
+            self.file.write('\n')
             indent += 1
             for key, value in item.items():
                 self.write_scalar(key, indent, pad=pad)
@@ -1050,8 +1181,10 @@ class _Writer:
         if comment is not None:
             self.file.write(f'#<{escape(comment)}> ')
         self.file.write(f'<{escape(item.name)}>')
-        for name in item.fieldnames:
-            self.file.write(f' <{escape(name)}>')
+        for field in item.fields:
+            self.file.write(f' <{escape(field.name)}>')
+            if field.vtype is not None:
+                self.file.write(f' {field.vtype}')
         if len(item) == 0:
             self.file.write(' = =]\n')
         else:
@@ -1080,7 +1213,7 @@ class _Writer:
         elif isinstance(item, int):
             self.file.write(str(item))
         elif isinstance(item, float):
-            self.file.write(_realstr(item))
+            self.file.write(realize(item))
         elif isinstance(item, (datetime.date, datetime.datetime)):
             self.file.write(item.isoformat())
         elif isinstance(item, str):
@@ -1097,7 +1230,7 @@ class _Writer:
                 raise Error('can only convert complex to ntuple if '
                             'one_way_conversion is True')
             self.file.write(
-                f'(:{_realstr(item.real)} {_realstr(item.imag)}:)')
+                f'(:{realize(item.real)} {realize(item.imag)}:)')
         elif isinstance(item, NTuple):
             self.file.write(item.asuxf)
         else:
@@ -1106,8 +1239,13 @@ class _Writer:
         return False
 
 
-def _realstr(s):
-    value = str(s)
+def realize(n: float) -> str:
+    '''Returns a str representation of the given number n with the str
+    guaranteed to include a decimal point and a digit either side of the
+    point whether the representation is standard or scientific, as required
+    by the UXF BNF.
+    '''
+    value = str(n)
     if '.' not in value:
         i = value.find('e')
         if i == -1:
@@ -1125,13 +1263,19 @@ def _is_scalar(x):
             bytearray))
 
 
+def _table_type_for_name(name):
+    return dict(int=int, date=datetime.date, datetime=datetime.datetime,
+                str=str, bytes=(bytes, bytearray), bool=bool,
+                real=float, uxf=None).get(name)
+
+
 def naturalize(s):
     '''Given string s returns True if the string is 't', 'true', 'y', 'yes',
     or False if the string is 'f', 'false', 'n', 'no' (case-insensitive), or
-    a datetime.datetime if s holds a parsable ISO8601 datetime string, or a
-    datetime.date if s holds a parsable ISO8601 date string, or an int if s
-    holds a parsable int, or a float if s holds a parsable float, or failing
-    these returns the string s unchanged.
+    an int if s holds a parsable int, or a float if s holds a parsable
+    float, or a datetime.datetime if s holds a parsable ISO8601 datetime
+    string, or a datetime.date if s holds a parsable ISO8601 date string, or
+    failing these returns the string s unchanged.
     '''
     u = s.upper()
     if u in {'T', 'TRUE', 'Y', 'YES'}:
@@ -1139,20 +1283,20 @@ def naturalize(s):
     if u in {'F', 'FALSE', 'N', 'NO'}:
         return False
     try:
-        if 'T' in s:
-            if isoparse is not None:
-                return isoparse(s)
-            return datetime.datetime.fromisoformat(s)
-        else:
-            if isoparse is not None:
-                return isoparse(s).date()
-            return datetime.date.fromisoformat(s)
+        return int(s)
     except ValueError:
         try:
-            return int(s)
+            return float(s)
         except ValueError:
             try:
-                return float(s)
+                if 'T' in s:
+                    if isoparse is not None:
+                        return isoparse(s)
+                    return datetime.datetime.fromisoformat(s)
+                else:
+                    if isoparse is not None:
+                        return isoparse(s).date()
+                    return datetime.date.fromisoformat(s)
             except ValueError:
                 return s
 
