@@ -3,17 +3,23 @@
 # License: GPLv3
 
 import argparse
+import configparser
 import csv
 import datetime
 import json
 import pathlib
+import sqlite3
+import sys
 
 import uxf
 
 
 def main():
     config = _get_config()
-    config.convert(config)
+    try:
+        config.convert(config)
+    except (FileNotFoundError, uxf.Error) as err:
+        print(err, file=sys.stderr)
 
 
 def _get_config():
@@ -94,7 +100,7 @@ def uxf_to_csv(config):
     if isinstance(data, uxf.Table):
         with open(config.outfile, 'w') as file:
             writer = csv.writer(file, quoting=csv.QUOTE_NONNUMERIC)
-            writer.writerow(data.fieldnames)
+            writer.writerow((field.name for field in data.fields))
             for row in data:
                 writer.writerow(row)
     elif (isinstance(data, (list, uxf.List)) and data and
@@ -125,7 +131,8 @@ def _read_csv_to_data(config):
             if data is None:
                 if config.fieldnames:
                     data = uxf.Table(name=pathlib.Path(filename).stem,
-                                     fieldnames=list(row))
+                                     fields=[uxf.Field(name)
+                                             for name in row])
                     continue
                 else:
                     data = []
@@ -165,23 +172,56 @@ class _JsonEncoder(json.JSONEncoder):
             return {JSON_BYTES: obj.hex().upper()}
         if isinstance(obj, (list, uxf.List)):
             comment = getattr(obj, COMMENT, None)
-            if comment is not None:
-                return {JSON_LIST: dict(comment=comment, list=list(obj))}
-            return list(obj)
-        if isinstance(obj, (dict, uxf.Map)):
-            comment = getattr(obj, COMMENT, None)
-            if comment is not None:
-                return {JSON_MAP: dict(comment=comment, map=dict(obj))}
-            return dict(obj)
+            vtype = getattr(obj, VTYPE, None)
+            if comment is None and vtype is None:
+                return list(obj)
+            return {JSON_LIST: dict(comment=comment, vtype=vtype,
+                                    list=list(obj))}
+        if isinstance(obj, uxf.Map):
+            return _json_encode_map(obj)
+        if isinstance(obj, dict):
+            return obj
         if isinstance(obj, uxf.NTuple):
             return {JSON_NTUPLE: obj.astuple}
         if isinstance(obj, uxf.Table):
             return {JSON_TABLE: dict(
-                comment=obj.comment, name=obj.name,
-                fieldnames=obj.fieldnames, records=obj.records)}
+                name=obj.name, comment=obj.comment,
+                fields={field.name: field.vtype for field in obj.fields},
+                records=obj.records)}
         return json.JSONEncoder.default(self, obj)
 
 
+def _json_encode_map(obj):
+    comment = getattr(obj, COMMENT, None)
+    ktype = getattr(obj, KTYPE, None)
+    vtype = getattr(obj, VTYPE, None)
+    d = {}
+    itypes = {}
+    for key, value in obj.items():
+        if isinstance(key, (datetime.date, datetime.datetime)):
+            skey = key.isoformat()
+            itypes[skey] = UXF
+        elif isinstance(key, int):
+            skey = str(key)
+            itypes[skey] = UXF
+        elif isinstance(key, (bytes, bytearray)):
+            skey = key.hex().upper()
+            itypes[skey] = BYTES
+        elif isinstance(key, str):
+            skey = key
+        else:
+            raise SystemExit(f'invalid map key type: {key} of {type(key)}')
+        d[skey] = value
+    if not itypes and comment is None and ktype is None:
+        return dict(obj)
+    m = dict(comment=comment, ktype=ktype, vtype=vtype, map=d)
+    if len(itypes) == len(d) and len(set(itypes.values())) == 1:
+        m[ITYPE] = itypes.popitem()[1] # all use same non-str key
+    elif itypes:
+        m[ITYPES] = itypes
+    return {JSON_MAP: m}
+
+
 def json_to_uxf(config):
     filename = config.infiles[0]
     with open(filename, 'rt', encoding=UTF8) as file:
@@ -197,38 +237,110 @@ def _json_naturalize(d):
     if JSON_BYTES in d:
         return bytes.fromhex(d[JSON_BYTES])
     if JSON_LIST in d:
-        x = uxf.List(d[JSON_LIST]['list'])
-        x.comment = d[JSON_LIST][COMMENT]
-        return x
+        jlist = d[JSON_LIST]
+        ls = uxf.List(jlist[LIST])
+        ls.comment = jlist[COMMENT]
+        ls.vtype = jlist[VTYPE]
+        return ls
     if JSON_MAP in d:
-        x = uxf.Map(d[JSON_MAP]['map'])
-        x.comment = d[JSON_MAP][COMMENT]
-        return x
+        jmap = d[JSON_MAP]
+        itype = jmap.get(ITYPE) # str or None
+        itypes = jmap.get(ITYPES) # dict or None
+        m = uxf.Map()
+        m.comment = jmap[COMMENT]
+        m.ktype = jmap[KTYPE] # str or None
+        m.vtype = jmap[VTYPE] # str or None
+        for key, value in jmap[MAP].items():
+            if itypes is not None:
+                itype = itypes.get(key)
+            if itype == BYTES:
+                key = bytes.fromhex(key)
+            else:
+                key = key if itype is None else uxf.naturalize(key)
+            m[key] = value
+        return m
     elif JSON_NTUPLE in d:
         return uxf.NTuple(*d[JSON_NTUPLE])
     elif JSON_TABLE in d:
-        return uxf.Table(name=d[JSON_TABLE]['name'],
-                         fieldnames=d[JSON_TABLE]['fieldnames'],
-                         records=d[JSON_TABLE]['records'],
-                         comment=d[JSON_TABLE][COMMENT])
+        jtable = d[JSON_TABLE]
+        fields = [uxf.Field(name, vtype) for name, vtype in
+                  jtable[FIELDS].items()]
+        return uxf.Table(name=jtable[NAME], comment=jtable[COMMENT],
+                         fields=fields, records=jtable[RECORDS])
     return d
 
 
-JSON_DATETIME = 'UXF^datetime'
-JSON_DATE = 'UXF^date'
-JSON_BYTES = 'UXF^bytes'
-JSON_LIST = 'UXF^list'
-JSON_MAP = 'UXF^map'
-JSON_NTUPLE = 'UXF^ntuple'
-JSON_TABLE = 'UXF^table'
-
-
 def ini_to_uxf(config):
-    print('ini_to_uxf', config) # TODO
+    ini = configparser.ConfigParser()
+    filename = config.infiles[0]
+    ini.read(filename)
+    data = uxf.Map()
+    for section in ini:
+        d = ini[section]
+        if d:
+            m = data[section] = uxf.Map()
+            for key, value in d.items():
+                m[uxf.naturalize(key)] = uxf.naturalize(value)
+    uxf.dump(config.outfile, data=data, custom=filename,
+             one_way_conversion=True)
 
 
 def uxf_to_sqlite(config):
-    print('uxf_to_sqlite', config) # TODO
+    data, _ = uxf.load(config.infiles[0])
+    if isinstance(data, uxf.Table):
+        _uxf_to_sqlite(config, [data])
+    elif (isinstance(data, (list, uxf.List)) and data and
+            all(isinstance(v, uxf.Table) for v in data)):
+        _uxf_to_sqlite(config, data)
+    else:
+        raise SystemExit('can only convert a UXF containing a single table '
+                         'or a single list of Tables to SQLite')
+
+
+def _uxf_to_sqlite(config, tables):
+    sqlite3.register_adapter(bool, lambda b: 'TRUE' if b else 'FALSE')
+    sqlite3.register_adapter(datetime.date, lambda d: d.isoformat())
+    sqlite3.register_adapter(datetime.datetime, lambda d: d.isoformat())
+    db = None
+    try:
+        db = sqlite3.connect(config.outfile)
+        for table_index, table in enumerate(tables, 1):
+            table_name = _create_table(db, table, table_index)
+            _populate_table(db, table, table_name)
+    finally:
+        if db is not None:
+            db.commit()
+            db.close()
+
+
+def _create_table(db, table, table_index):
+    table_name = uxf._canonicalize(table.name, f'Table{table_index}')
+    sql = ['CREATE TABLE IF NOT EXISTS ', table_name, ' (']
+    types = ['TEXT'] * len(table.fieldnames)
+    if table.records:
+        for i, value in enumerate(table.records[0]):
+            if isinstance(value, float):
+                types[i] = 'REAL'
+            elif isinstance(value, int) and not isinstance(value, bool):
+                types[i] = 'INT'
+            elif isinstance(value, (bytes, bytearray)):
+                types[i] = 'BLOB'
+    sep = ''
+    for i, name in enumerate(table.fieldnames):
+        field_name = uxf._canonicalize(name, f'Field{i + 1}')
+        sql += [sep, field_name, ' ', types[i]]
+        sep = ', '
+    sql += [');']
+    cursor = db.cursor()
+    cursor.execute(''.join(sql))
+    return table_name
+
+
+def _populate_table(db, table, table_name):
+    sql = ['INSERT INTO ', table_name, ' VALUES (',
+           ', '.join('?' * len(table.fieldnames)), ');']
+    cursor = db.cursor()
+    cursor.executemany(''.join(sql), table.records)
 
 
 def sqlite_to_uxf(config):
@@ -243,6 +355,8 @@ def xml_to_uxf(config):
     print('xml_to_uxf', config) # TODO
 
 
+BYTES = 'bytes'
+COMMENT = 'comment'
 DOT_CSV = '.CSV'
 DOT_INI = '.INI'
 DOT_JSN = '.JSN'
@@ -250,8 +364,24 @@ DOT_JSON = '.JSON'
 DOT_SQLITE = '.SQLITE'
 DOT_UXF = '.UXF'
 DOT_XML = '.XML'
-COMMENT = 'comment'
+FIELDS = 'fields'
+ITYPE = 'itype'
+ITYPES = 'itypes'
+JSON_BYTES = 'UXF^bytes'
+JSON_DATE = 'UXF^date'
+JSON_DATETIME = 'UXF^datetime'
+JSON_LIST = 'UXF^list'
+JSON_MAP = 'UXF^map'
+JSON_NTUPLE = 'UXF^ntuple'
+JSON_TABLE = 'UXF^table'
+KTYPE = 'ktype'
+LIST = 'list'
+MAP = 'map'
+NAME = 'name'
+RECORDS = 'records'
 UTF8 = 'utf-8'
+UXF = 'uxf'
+VTYPE = 'vtype'
 
 USAGE = '''
 uxfconvert.py <infile.uxf> <outfile.{csv,json,sqlite,xml}>
@@ -276,6 +406,10 @@ scalars otherwise.
 Converting from uxf to csv can only be done if the uxf contains a single
 table or a single list of lists of scalars.
 
+Converting from ini to uxf is purely for example purposes (and drops ini
+comments). In a real application (e.g., migrating from ini to uxf), a custom
+ini parser would be needed.
+
 Converting from uxf to json and back (i.e., using uxfconvert.py's own json
 format) should work with perfect fidelity.
 
